module FLouvainSpec where

import Test.Hspec
import Test.HUnit.Approx (assertApproxEqual)

import Protolude

-- FGL
import Data.Graph.Inductive
import qualified Data.Graph.Inductive as DGI
import Data.List ((!!))

import Data.Graph.Clustering.FLouvain
import Data.Graph.Clustering.Louvain.Utils (mkFGraph, mkFGraph')
import Data.Graph.Clustering.Louvain.Types
import Data.Graph.FGL

-- 1 -> 2 -> 3

simpleGraph :: FGraph () ()
simpleGraph = mkFGraph' [ (1, 2, 1.0)
                        , (2, 3, 0.5)
                        ]

simpleLGraph :: FGraph Text ()
simpleLGraph = mkFGraph [ (1, "one")
                        , (2, "two")
                        , (3, "three")]
                        [ (1, 2, 1.0)
                        , (2, 3, 0.5) ]

spec :: Spec
spec = do
  describe "FLouvain tests" $ do
    it "graphWeight computes correctly" $ do
      assertApproxEqual "graphWeights don't match"
                        0.00001
                        1.5
                        (unGraphWeightSum $ graphWeight simpleGraph)

    it "nodeWeightSum computes correctly" $ do
      nodeWeightSum (DGI.context simpleGraph 1) `shouldBe` NodeWeightSum 1.0
      nodeWeightSum (DGI.context simpleGraph 2) `shouldBe` NodeWeightSum 1.5
      nodeWeightSum (DGI.context simpleGraph 3) `shouldBe` NodeWeightSum 0.5

    it "replaceLNode works correctly" $ do
      let replaced = replaceLNode simpleLGraph (1, "ONE")
      nodes replaced `shouldBe` [1, 2, 3]
      lnodes replaced `shouldBe` ["ONE", "two", "three"]

    it "initialCgr computes correctly" $ do
      let cgr = initialCGr simpleGraph
          communities = lnodes cgr
          iws0 = InWeightSum 0.0
      nodes cgr `shouldBe` [1, 2, 3]
      edges cgr `shouldBe` [ (1, 2)
                           , (2, 3) ]
      Protolude.map comNodes communities `shouldBe` [[1], [2], [3]]
      Protolude.map comInWeightSum communities `shouldBe` [iws0, iws0, iws0]
      Protolude.map comTotWeightSum communities `shouldBe`
        [ TotWeightSum 1.0
        , TotWeightSum 1.5
        , TotWeightSum 0.5 ]

    it "nodeComWeightSum computes correctly" $ do
      let cgr = initialCGr simpleGraph
          communities = lnodes cgr
          fstCom = communities !! 0
          sndCom = communities !! 1
          trdCom = communities !! 2
      nodeComWeightSum fstCom (DGI.context simpleGraph 1) `shouldBe` NodeComWeightSum 0.0
      nodeComWeightSum fstCom (DGI.context simpleGraph 2) `shouldBe` NodeComWeightSum 1.0
      nodeComWeightSum fstCom (DGI.context simpleGraph 3) `shouldBe` NodeComWeightSum 0.0
      nodeComWeightSum sndCom (DGI.context simpleGraph 1) `shouldBe` NodeComWeightSum 1.0
      nodeComWeightSum sndCom (DGI.context simpleGraph 2) `shouldBe` NodeComWeightSum 0.0
      nodeComWeightSum sndCom (DGI.context simpleGraph 3) `shouldBe` NodeComWeightSum 0.5
      nodeComWeightSum trdCom (DGI.context simpleGraph 1) `shouldBe` NodeComWeightSum 0.0
      nodeComWeightSum trdCom (DGI.context simpleGraph 2) `shouldBe` NodeComWeightSum 0.5
      nodeComWeightSum trdCom (DGI.context simpleGraph 3) `shouldBe` NodeComWeightSum 0.0

    it "modularity computes correctly" $ do
      let cgr = initialCGr simpleGraph
          m = graphWeight simpleGraph
          mod = modularity simpleGraph cgr m
          c1 = 0.0 - (1.0 * 1.0) / (2.0 * 3.0)
          c2 = 0.0 - (3.0 * 3.0) / (2.0 * 3.0)
          c3 = 0.0 - (2.0 * 2.0) / (2.0 * 3.0)
          mMod = (c1 + c2 + c3) / (2.0 * 3.0)
      assertApproxEqual "modularities don't match"
                        0.000001
                        mMod
                        (unModularity mod)

    it "tests that single community has modularity 0" $ do
      let cgr = mkGraph [
            (1, Community(nodes simpleGraph, InWeightSum $ sumEdgeWeights $ edges simpleGraph, TotWeightSum 0.0, ()))
            ] []
          m = graphWeight simpleGraph
          mod = modularity simpleGraph cgr m
      assertApproxEqual "modularities don't match"
                        0.000001
                        0.0
                        (unModularity mod)

    it "nodeCommunity works correctly" $ do
      let cgr = initialCGr simpleGraph
          communities = lnodes cgr
      (nodeCommunity 1 cgr) `shouldBe` Just (1, communities !! 0)
      (nodeCommunity 2 cgr) `shouldBe` Just (2, communities !! 1)
      (nodeCommunity 3 cgr) `shouldBe` Just (3, communities !! 2)

    it "nodeNeighbours works correctly" $ do
      let cgr = initialCGr simpleGraph
          communities = lnodes cgr
      (nodeNeighbours 1 cgr) `shouldBe` [(2, communities !! 1)]
      (nodeNeighbours 2 cgr) `shouldBe` [(1, communities !! 0), (3, communities !! 2)]
      (nodeNeighbours 3 cgr) `shouldBe` [(2, communities !! 1)]

    it "moveNodeWithContext works correctly" $ do
      let cgr = initialCGr simpleGraph
          communities = lnodes cgr
          ctx1 = DGI.context simpleGraph 1
          ctx2 = DGI.context simpleGraph 2
          ctx3 = DGI.context simpleGraph 3
          com1 = communities !! 0
          com2 = communities !! 1
          newCom1 = moveNodeWithContext ctx1 OutOf com1
          newCom2 = moveNodeWithContext ctx1 Into com2
          intoOutOf ctx com = moveNodeWithContext ctx OutOf $ moveNodeWithContext ctx Into com
          outOfInto ctx com = moveNodeWithContext ctx Into $ moveNodeWithContext ctx OutOf com
      newCom1 `shouldBe` Community ([], InWeightSum 0.0, TotWeightSum 0.0, comLabel newCom1)
      newCom2 `shouldBe` Community ([1, 2], InWeightSum 1.0, TotWeightSum 0.5, comLabel newCom1)

      -- TODO moveNodeWithContext ctx Into (moveNodeWithContext ctx OutOf) is an
      -- identity, this can be used in QuickCheck testing
      -- Same thing the other way (OutOf . Into)
      intoOutOf ctx1 com2 `shouldBe` com2
      outOfInto ctx1 com1 `shouldBe` com1

      intoOutOf ctx3 newCom2 `shouldBe` newCom2
      outOfInto ctx2 newCom2 `shouldBe` newCom2
